use LWP::UserAgent;

our $pogl = $ARGV[0] eq 'pogl';
our $tw = $pogl ? 100 : 90;
our $docs = $pogl ?
  'S:/gcWeb/gcSite/w01/s3_Documentation/data/opengl.html' :
  'w:/www/opengl/apis.html';

our $dest = 'w:/www/info/specs/opengl/';
our $ver = "1.0";
our $vendors = {};
our $funcs = {};
our $ext = {};

$ext->{'1.0'}->{url} = 'http://www.glprogramming.com/blue';
$ext->{'1.1'}->{url} = 'http://opengl.org/documentation/specs/version1.1/glspec1.1';
$ext->{'1.2'}->{url} = 'http://opengl.org/documentation/specs/version1.2/opengl1.2.1.pdf';
$ext->{'1.3'}->{url} = 'http://opengl.org/documentation/specs/version1.3/glspec13.pdf';
$ext->{'1.4'}->{url} = 'http://opengl.org/documentation/specs/version1.4/glspec14.pdf';
$ext->{'1.5'}->{url} = 'http://opengl.org/documentation/specs/version1.5/glspec15.pdf';
$ext->{'2.0'}->{url} = 'http://opengl.org/documentation/specs/version2.0/glspec20.pdf';
$ext->{'2.1'}->{url} = 'http://www.opengl.org/registry/doc/glspec21.20061201.pdf';


my $urls = 'api_urls.txt';
die "Unable to read '$urls'" if (!open(URLS,$urls));
print "Parsing $urls\n";

while (<URLS>)
{
  my $line = $_;
  $line =~ s|[\r\n]+||g;

  my($api,$url) = split(',',$line);
  next if (!$url);

  $funcs->{$api}->{url} = $url;
  #print "Have URL for: $api\n";
}
close(URLS);
die "Unable to append to '$urls'" if (!open(URLS,">>$urls"));


my $exts = 'ext_urls.txt';
die "Unable to read '$exts'" if (!open(EXTS,$exts));
print "Parsing $exts\n";

while (<EXTS>)
{
  my $line = $_;
  $line =~ s|[\r\n]+||g;

  my($ver,$url) = split(',',$line);
  next if (!$url);

  $ext->{$ver}->{url} = $url;
  #print "Have URL for: $ver\n";
}
close(EXTS);
die "Unable to append to '$exts'" if (!open(EXTS,">>$exts"));


# Parse gl.h
my $gl = '../include/GL/gl.h';
die "Unable to read '$gl'" if (!open(GL,$gl));
print "Parsing $gl\n";

# Skip to procs
while (<GL>)
{
  my $line = $_;
  last if ($line =~ m|GL_TEXTURE_COMPONENTS|);
}

# Get APIs
while (<GL>)
{
  my $line = $_;
  last if ($line =~ m|^\#ifdef __cplusplus|);

  if ($line =~ m|APIENTRY ([^ ]+)|)
  {
    my $api = $1;
    $funcs->{$api}->{ver}=$ver;

    next if ($funcs->{$api}->{url});
    my $url = GetURL($api);
    $funcs->{$api}->{url}=$url;
    print URLS "$api,$url\n";
  }
};
close(GL);



# Parse glext.h
my $glext = '../include/GL/glext.h';
die "Unable to read '$glext'" if (!open(GLEXT,$glext));
print "Parsing $glext\n";

# Skip to procs
while (<GLEXT>)
{
  my $line = $_;
  last if ($line =~ m|^\#include \<stddef.h\>|);
}

# Handle extensions
while (<GLEXT>)
{
  my $line = $_;
  last if ($line =~ m|^\#ifdef __cplusplus|);

  # Get current version/extension
  if ($line =~ m|^\#define GL_VERSION_([^ ]+)|)
  {
    $ver = $1;
    $ver =~ s|_|.|g;
    next;
  }
  elsif ($line =~ m|^\#define (GL_[^ ]+)|)
  {
    $ver = $1;
    $ver =~ m|GL_([^_]+)_(.*)|;
    my $vendor = $1;
    my $name = $2;
    my $url = "http://opengl.org/registry/specs/$vendor/$name.txt";

    $vendors->{$vendor}++;
    next if ($ext->{$ver}->{url});

    print "Ext: $url\n";
    my $txt_file = $ver.'.txt';
    if (!Have($txt_file))
    {
      my $txt_doc = Fetch($url);
      if ($txt_doc)
      {
        Save($txt_file,$txt_doc);
      }
      else
      {
        $url= 'NONE';
      }
    }
    print EXTS "$ver,$url\n";

    $ext->{$ver}->{url} = $url;
    next;
  }

  if ($line =~ m|APIENTRY ([^ ]+)|)
  {
    my $api = $1;
    next if ($api =~ m|^PFN|);

    $funcs->{$api}->{ver}=$ver;

    next if ($funcs->{$api}->{url});
    my $url = GetURL($api);
    $funcs->{$api}->{url}=$url;
    print URLS "$api,$url\n";
  }
};
close(GLEXT);

close(EXTS);
close(URLS);


# Parse OpenGL.xs
my $xs = '../OpenGL.xs';
die "Unable to read '$xs'" if (!open(XS,$xs));
print "Parsing $xs\n";

# Get APIs
my $n = 0;
my $last;
while (<XS>)
{
  my $line = $_;
  $n++;

  if ($line =~ m|^\#//\# (.*;)|)
  {
    my $doc = $1;

    $doc =~ m|([^ \(]+)\(|;
    my $fnc = $1;

    my $api = $fnc;
    my $type = ($api =~ s/_(c|s|p)$//) ? $1 : '';

    if (!defined($funcs->{$api}))
    {
      #print "Not found: $api - line $n\n";
      #my $url = GetURL($api);
      #$funcs->{$api}->{url}=$url;
    }
    $funcs->{$api}->{xs}++;

    if ($type)
    {
      $funcs->{$api}->{"doc_$type"} = $doc;
    }
    else
    {
      $funcs->{$api}->{doc} = $doc;
    }

    $last = $fnc;
  }
  elsif ($line =~ m|^\#//- (.*)|)
  {
    $note = $1;
    $note =~ s|[\r\n]+||g;

    my $api = $last;
    my $type = ($api =~ s/_(c|s|p)$//) ? $1 : '';

    if ($type)
    {
      $funcs->{$api}->{"note_$type"} = $note;
    }
    else
    {
      $funcs->{$api}->{note} = $note;
    }
  }
};
close(XS);




print "##########\n";
my $out;
if (!$pogl)
{
$out = qq
{<html>
<head>
<title>POGL - OpenGL</title>
<meta name="description" content="POGL - Portable OpenGL APIs">
<meta name="keywords" content="POGL,Documentation,OpenGL,POGL,Perl OpenGL,POGL APIs,Perl OpenGL APIs,Portable">
<meta http-equiv="imagetoolbar" content="no">
</head>

<body bgcolor=#F0F0F0>
<center>

<script type="text/javascript"><!--
google_ad_client = "pub-8693589755201542";
google_ad_width = 728;
google_ad_height = 90;
google_ad_format = "728x90_as";
google_ad_type = "text";
//2007-04-03: Video Leaderboard
google_ad_channel = "0160174477";
google_color_border = "F0F0F0";
google_color_bg = "F0F0F0";
google_color_link = "00AAFF";
google_color_text = "000000";
google_color_url = "005588";
//-->
</script>
<script type="text/javascript"
  src="http://pagead2.googlesyndication.com/pagead/show_ads.js">
</script>
<br/>
<a href="https://adwords.google.com/select/OnsiteSignupLandingPage?client=ca-pub-8693589755201542&referringUrl=http://graphcomp.com/">Advertise on this site</a>

<hr/>
<table width="100%">
<tr>
<td width="20%" align="left">
<script type="text/javascript"><!--
google_ad_client = "pub-8693589755201542";
/* 125x125, created 5/8/08 */
google_ad_slot = "4971375492";
google_ad_width = 125;
google_ad_height = 125;
google_cpa_choice = ""; // on file
//-->
</script>
<script type="text/javascript"
src="http://pagead2.googlesyndication.com/pagead/show_ads.js">
</script>
</td>

<td width="60%" align="center">
<center>
Brought to you by the<br/>
<a href="/info/specs"><img src="/info/images/specsbar.gif" width="360" height="60" border="0"></a>
<br/>
<!-- SiteSearch Google -->
<table>
<form method="get" action="http://www.google.com/custom" target="google_window">
<tr>
<td>
<input type="hidden" name="domains" value="graphcomp.com"></input>
<label for="sbi" style="display: none">Enter your search terms</label>
<input type="text" name="q" size="30" maxlength="255" value="OpenGL" id="sbi"></input>
<label for="sbb" style="display: none">Submit search form</label>
<input type="submit" name="sa" value="Search" id="sbb"></input>
</td>
</tr>
<tr>
<td>
<input type="radio" name="sitesearch" value="graphcomp.com" checked id="ss1"></input>
<label for="ss1" title="Search graphcomp.com"><font size="-1" color="#000000">graphcomp.com</font></label>
<input type="radio" name="sitesearch" value="" id="ss0"></input>
<label for="ss0" title="Search the Web"><font size="-1" color="#000000">Web</font></label>
<!--<a href="http://www.google.com/"><img src="http://www.google.com/logos/Logo_25gry.gif" border="0" alt="Google" align="middle"></img></a>-->
<input type="hidden" name="client" value="pub-8693589755201542"></input>
<input type="hidden" name="forid" value="1"></input>
<input type="hidden" name="channel" value="6059963133"></input>
<input type="hidden" name="ie" value="ISO-8859-1"></input>
<input type="hidden" name="oe" value="ISO-8859-1"></input>
<input type="hidden" name="safe" value="active"></input>
<input type="hidden" name="flav" value="0001"></input>
<input type="hidden" name="sig" value="6mMb45Nn9EH849Hp"></input>
<input type="hidden" name="cof" value="GALT:#0066CC;GL:1;DIV:#33FFFF;VLC:004488;AH:center;BGC:FFFFFF;LBGC:CCCCCC;ALC:0088FF;LC:0088FF;T:000000;GFNT:004488;GIMP:004488;LH:50;LW:300;L:http://www.graphcomp.com/info/images/specsbar.gif;S:http://graphcomp.com/info/specs;FORID:1"></input>
<input type="hidden" name="hl" value="en"></input>
</td>
</tr>
</form>
</table>
<!-- SiteSearch Google -->
</center>
</td>

<td width="20%" align="right">
<script type="text/javascript"><!--
google_ad_client = "pub-8693589755201542";
google_ad_width = 125;
google_ad_height = 125;
google_ad_format = "125x125_as_rimg";
google_cpa_choice = "CAAQjf6XhAIaCP7y-hV9rXWvKK2293M";
google_ad_channel = "6091888669";
//-->
</script>
<script type="text/javascript" src="http://pagead2.googlesyndication.com/pagead/show_ads.js">
</script>
</td>
</tr>
</table>

<hr/>
<h2><a href="/opengl">POGL</a>: Portable OpenGL APIs</h2>
</center>
};
}

$out .= qq
{<center>
<table width="$tw%">
<tr>
<td>
};

#$out .= "<hr/>\n" if (!$pogl);

$out .= qq
{
POGL provides interfaces to support OpenGL APIs on languages that do not support
C pointers.  This is handled via suffixed APIs that support native-language references:

<ul>
<code>_c</code> APIs provides <code>(CPTR)</code> interfaces for languages (eg: Perl) that provide a means
to pass around C pointers.
</ul>

<ul>
<code>_s</code> APIs provides <code>(PACKED)</code> string interfaces for most other languages.
</ul>

<ul>
<code>_p</code> are portable APIs that do not require passing pointers; instead, blocks of data
may be passed around via <code>(OGA)</code>
<a href="/opengl/arrays">OpenGL::Array</a> object references.
</ul>
};

foreach $api (sort(keys(%$funcs)))
{
  next if ($pogl && !$funcs->{$api}->{xs});
  next if ($api !~ m|^gl|);
  next if ($api =~ m|^glp|);
  next if ($api =~ m|^glu|);
  print "$api\n";

  $out .= "<hr>\n";

  $out .= qq
  {
<table>
<tr valign="top">
<td>
<!-- SiteSearch Google -->
<form method="get" action="http://www.google.com/custom" target="google_window">
<input type="hidden" name="domains" value="graphcomp.com"></input>
<!--<label for="sbi" style="display: none">Enter your search terms</label>-->
<input type="hidden" name="q" size="30" maxlength="255" value="$api" id="sbi"></input>
<!--<label for="sbb" style="display: none">Submit search form</label>-->
<input type="submit" name="sa" value="Google it" id="sbb"></input>
<!--<input type="radio" name="sitesearch" value="graphcomp.com" checked id="ss1"></input>-->
<!--<label for="ss1" title="Search graphcomp.com"><font size="-1" color="#000000">graphcomp.com</font></label>-->
<input type="hidden" name="sitesearch" value="" checked id="ss0"></input>
<!--<label for="ss0" title="Search the Web"><font size="-1" color="#000000">Web</font></label>-->
<!--<a href="http://www.google.com/"><img src="http://www.google.com/logos/Logo_25gry.gif" border="0" alt="Google" align="middle"></img></a>-->
<input type="hidden" name="client" value="pub-8693589755201542"></input>
<input type="hidden" name="forid" value="1"></input>
<input type="hidden" name="channel" value="6059963133"></input>
<input type="hidden" name="ie" value="ISO-8859-1"></input>
<input type="hidden" name="oe" value="ISO-8859-1"></input>
<input type="hidden" name="safe" value="active"></input>
<input type="hidden" name="flav" value="0001"></input>
<input type="hidden" name="sig" value="6mMb45Nn9EH849Hp"></input>
<input type="hidden" name="cof" value="GALT:#0066CC;GL:1;DIV:#33FFFF;VLC:004488;AH:center;BGC:FFFFFF;LBGC:CCCCCC;ALC:0088FF;LC:0088FF;T:000000;GFNT:004488;GIMP:004488;LH:50;LW:300;L:http://www.graphcomp.com/info/images/specsbar.gif;S:http://graphcomp.com/info/specs;FORID:1"></input>
<input type="hidden" name="hl" value="en"></input>
</form>
<!-- SiteSearch Google -->
</td>
<td>
  };

  $out .= "&nbsp;<b><font size=\"+1\">";
  my $api_url = $funcs->{$api}->{url};
  $out .= ($api_url && ($api_url ne 'NONE')) ? "<a href=\"$api_url\" target=\"pogl_apis\">$api</a>" : $api;
  $out .= "</font></b>";

  my $ver = $funcs->{$api}->{ver};
  if ($ver)
  {
    $out .= " - ";
    my $ver_url = $ext->{$ver}->{url};
    $out .= ($ver_url && ($ver_url ne 'NONE')) ? "<a href=\"$ver_url\" target=\"pogl_apis\">$ver</a>" : $ver;
  }

  $out .= qq
  {
</td>
</tr>
  };

  if ($funcs->{$api}->{doc})
  {
    $out .= "<tr><td></td><td>\n";
    $out .= "<code>$funcs->{$api}->{doc}</code>\n";
    $out .= "<br/>\n$funcs->{$api}->{note}\n" if ($funcs->{$api}->{note});
    $out .= "<p/></td></tr>\n";
  }

  if ($funcs->{$api}->{doc_c})
  {
    $out .= "<tr><td></td><td>\n";
    $out .= "<code>$funcs->{$api}->{doc_c}</code>\n";
    $out .= "<br/>\n$funcs->{$api}->{note_c}\n" if ($funcs->{$api}->{note_c});
    $out .= "<p/></td></tr>\n";
  }

  if ($funcs->{$api}->{doc_s})
  {
    $out .= "<tr><td></td><td>\n";
    $out .= "<code>$funcs->{$api}->{doc_s}</code>\n";
    $out .= "<br/>\n$funcs->{$api}->{note_s}\n" if ($funcs->{$api}->{note_s});
    $out .= "<p/></td></tr>\n";
  }

  if ($funcs->{$api}->{doc_p})
  {
    $out .= "<tr><td></td><td>\n";
    $out .= "<code>$funcs->{$api}->{doc_p}</code>\n";
    $out .= "<br/>\n$funcs->{$api}->{note_p}\n" if ($funcs->{$api}->{note_p});
    $out .= "<p/></td></tr>\n";
  }

  $out .= "</table>\n";
}

$out .= qq
{
</td>
</tr>
</table>
</center>
};

if (!$pogl)
{
$out .= qq
{
<HR>
<center>
<script type="text/javascript"><!--
google_ad_client = "pub-8693589755201542";
google_ad_width = 728;
google_ad_height = 90;
google_ad_format = "728x90_as";
google_ad_type = "text";
//2007-04-05: MPEG Footer
google_ad_channel = "7223635951";
google_color_border = "F0F0F0";
google_color_bg = "F0F0F0";
google_color_link = "00AAFF";
google_color_text = "000000";
google_color_url = "005588";
//-->
</script>
<script type="text/javascript"
  src="http://pagead2.googlesyndication.com/pagead/show_ads.js">
</script>
</center>

</body>
</html>
};
}

die "Unable to append to '$docs'" if (!open(DOCS,">$docs"));
print DOCS $out;
close(DOCS);
exit 0;




sub GetURL
{
  my($api) = @_;
  my $orig = length($api);
  my $len = $orig;

  while ($len > 4)
  {
    my $xml = "http://www.opengl.org/sdk/docs/man/xhtml/$api.xml";

    my $xml_file = $api.'.html';
    return $xml if (Have($xml_file));
    my $xml_doc = Fetch($xml);
    Save($xml_file,$xml_doc) if ($xml_doc);

    $api =~ m|gl(.*)|;
    my $htm = 'http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/';
    $htm .= lc($1).'.html';

    my $htm_file = $api.'.htm';
    return $htm if (Have($htm_file));
    my $htm_doc = Fetch($htm);
    Save($htm_file,$htm_doc) if ($htm_doc);

    return $xml if ($xml_doc);
    return $htm if ($htm_doc);

    if ($len == $orig)
    {
      foreach my $vend (keys %$vendors)
      {
        if ($api =~ s|$vend$||)
        {
          $len = length($api);
          last;
        }
      }
    }

    $api = substr($api,0,--$len);
  }

  return 'NONE';
}

sub Fetch
{
  my($url) = @_;

  my $ua = LWP::UserAgent->new;
  return undef if (!$ua);

  my $req = HTTP::Request->new(GET => $url);
  return undef if (!$req);
  print "Fetching: $url\n";

  my $res = $ua->request($req);
  return undef if (!$res);
  return undef if (!$res->is_success);

  return $res->content;
}

sub Save
{
  my($file,$data) = @_;

  my $path = "$dest$file";
  return if (-e $path);

  return if (!open(FILE, ">$path"));
  print FILE $data;
  close(FILE);

  print "Saved: $path\n";
}

sub Have
{
  my($file) = @_;
  my $path = "$dest$file";
  print "Have: $file\n" if (-e $path);
  return -e $path;
}